import json
import weakref

try:
    from unittest import mock
except ImportError:
    import mock

import django
from django.test import TestCase, TransactionTestCase, override_settings
from django.urls import reverse

from django_ses import utils as ses_utils
from django_ses.inbound import BaseHandler
from django_ses.signals import (
    bounce_received,
    click_received,
    complaint_received,
    delivery_received,
    open_received,
    send_received,
)
from tests.mocks import (
    get_mock_bounce,
    get_mock_click,
    get_mock_complaint,
    get_mock_delivery,
    get_mock_open,
    get_mock_received_sns,
    get_mock_send,
)


class HandleBounceTestCase(TestCase):
    """
    Test the bounce web hook handler.
    """
    def setUp(self):
        self._old_bounce_receivers = bounce_received.receivers
        bounce_received.receivers = []

        self._old_complaint_receivers = complaint_received.receivers
        complaint_received.receivers = []

    def tearDown(self):
        bounce_received.receivers = self._old_bounce_receivers
        complaint_received.receivers = self._old_complaint_receivers

    def test_handle_bounce_event(self):
        """
        Test handling a normal bounce request.
        """
        req_mail_obj, req_bounce_obj, notification = get_mock_bounce("eventType")

        def _handler(sender, mail_obj, bounce_obj, raw_message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(req_mail_obj, mail_obj)
            self.assertEqual(req_bounce_obj, bounce_obj)
            self.assertEqual(raw_message, json.dumps(notification).encode())
        _handler.call_count = 0
        bounce_received.connect(_handler)

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = True

            response = self.client.post(reverse("django_ses_bounce"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(_handler.call_count, 1)

    def test_handle_bounce_notification(self):
        """
        Test handling a normal bounce request.
        """
        req_mail_obj, req_bounce_obj, notification = get_mock_bounce("notificationType")

        def _handler(sender, mail_obj, bounce_obj, raw_message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(req_mail_obj, mail_obj)
            self.assertEqual(req_bounce_obj, bounce_obj)
            self.assertEqual(raw_message, json.dumps(notification).encode())
        _handler.call_count = 0
        bounce_received.connect(_handler)

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = True

            response = self.client.post(reverse("django_ses_bounce"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(_handler.call_count, 1)


# We're using a TransactionTestCase instead of a TestCase because we want to
# avoid using transactions. Calls to atomic() will disconnect the signal
# handlers, which is exactly what we're trying to test here.
class HandleBounceTestCaseWithBL(TransactionTestCase):
    @override_settings(AWS_SES_VERIFY_BOUNCE_SIGNATURES=False)
    def test_bounce_event_can_perform_blacklist(self):
        """
        Test if bounce events result in blacklisted emails (if the feature has
        been enabled)
        """

        mail_obj, bounce_obj, notification = get_mock_bounce("eventType")

        # We can't use mock.patch("django_ses.signals.bounce_handler") here
        # because the signal gets registered way before this code is executed.
        # We would be late even if we tried to patch it in the setUp() method.
        # The only way is to patch the actual .receivers prop at "this" exact
        # moment.
        mock_func = mock.MagicMock()
        receiver = bounce_received.receivers[0]

        # Django<5 had only synchronous receivers, but >=5 can have asynchronous
        # ones as well. This caused the receivers structure to change:
        #
        # 4.2.16: https://github.com/django/django/blob/4.2.16/django/dispatch/dispatcher.py#L253
        # 5.0: https://github.com/django/django/blob/5.0/django/dispatch/dispatcher.py#L431
        #
        # We must account for that difference and prepare a tuple that will
        # match the expected signature.
        if django.VERSION[0] < 5:
            receiver = (receiver[0], weakref.ref(mock_func))
        else:
            receiver = (receiver[0], weakref.ref(mock_func), False)
        bounce_received.receivers[0] = receiver
        response = self.client.post(reverse("django_ses_bounce"), json.dumps(notification),
                                    content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(mock_func.call_count, 1)

    @override_settings(AWS_SES_VERIFY_BOUNCE_SIGNATURES=False)
    def test_complaint_event_can_perform_blacklist(self):
        """
        Test if complaint events result in blacklisted emails (if the feature
        has been enabled)
        """

        mail_obj, complaint_obj, notification = get_mock_complaint("eventType")

        # We can't use mock.patch("django_ses.signals.complaint_handler") here
        # because the signal gets registered way before this code is executed.
        # We would be late even if we tried to patch it in the setUp() method.
        # The only way is to patch the actual .receivers prop at "this" exact
        # moment.
        mock_func = mock.MagicMock()
        receiver = complaint_received.receivers[0]

        # Django<5 had only synchronous receivers, but >=5 can have asynchronous
        # ones as well. This caused the receivers structure to change:
        #
        # 4.2.16: https://github.com/django/django/blob/4.2.16/django/dispatch/dispatcher.py#L253
        # 5.0: https://github.com/django/django/blob/5.0/django/dispatch/dispatcher.py#L431
        #
        # We must account for that difference and prepare a tuple that will
        # match the expected signature.
        if django.VERSION[0] < 5:
            receiver = (receiver[0], weakref.ref(mock_func))
        else:
            receiver = (receiver[0], weakref.ref(mock_func), False)

        complaint_received.receivers[0] = receiver
        response = self.client.post(reverse("django_ses_bounce"), json.dumps(notification),
                                    content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(mock_func.call_count, 1)


class HandleEventTestCase(TestCase):
    """
    Test the event web hook handler.
    """
    def setUp(self):
        self._old_bounce_receivers = bounce_received.receivers
        bounce_received.receivers = []

        self._old_complaint_receivers = complaint_received.receivers
        complaint_received.receivers = []

    def tearDown(self):
        bounce_received.receivers = self._old_bounce_receivers
        complaint_received.receivers = self._old_complaint_receivers

    def test_handle_bounce_event(self):
        """
        Test handling a normal bounce request.
        """
        req_mail_obj, req_bounce_obj, notification = get_mock_bounce("eventType")

        def _handler(sender, mail_obj, bounce_obj, raw_message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(req_mail_obj, mail_obj)
            self.assertEqual(req_bounce_obj, bounce_obj)
            self.assertEqual(raw_message, json.dumps(notification).encode())
        _handler.call_count = 0
        bounce_received.connect(_handler)

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = True
            response = self.client.post(reverse("event_webhook"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(_handler.call_count, 1)


    def test_handle_bounce_notification(self):
        """
        Test handling a normal bounce request.
        """
        req_mail_obj, req_bounce_obj, notification = get_mock_bounce("notificationType")

        def _handler(sender, mail_obj, bounce_obj, raw_message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(req_mail_obj, mail_obj)
            self.assertEqual(req_bounce_obj, bounce_obj)
            self.assertEqual(raw_message, json.dumps(notification).encode())
        _handler.call_count = 0
        bounce_received.connect(_handler)

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = True
            response = self.client.post(reverse("event_webhook"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(_handler.call_count, 1)


    def test_handle_send_event(self):
        """
        Test handling a send event request.
        """
        req_mail_obj, req_send_obj, notification = get_mock_send()

        def _handler(sender, mail_obj, send_obj, raw_message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(req_mail_obj, mail_obj)
            self.assertEqual(req_send_obj, send_obj)
            self.assertEqual(raw_message, json.dumps(notification).encode())
        _handler.call_count = 0
        send_received.connect(_handler)

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = True
            response = self.client.post(reverse("event_webhook"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(_handler.call_count, 1)

    def test_handle_delivery_event(self):
        """
        Test handling a delivery event request.
        """
        req_mail_obj, req_delivery_obj, notification = get_mock_delivery()

        def _handler(sender, mail_obj, delivery_obj, raw_message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(req_mail_obj, mail_obj)
            self.assertEqual(req_delivery_obj, delivery_obj)
            self.assertEqual(raw_message, json.dumps(notification).encode())
        _handler.call_count = 0
        delivery_received.connect(_handler)

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = True
            response = self.client.post(reverse("event_webhook"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(_handler.call_count, 1)

    def test_handle_open_event(self):
        """
        Test handling a open event request.
        """
        req_mail_obj, req_open_obj, notification = get_mock_open()

        def _handler(sender, mail_obj, open_obj, raw_message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(req_mail_obj, mail_obj)
            self.assertEqual(req_open_obj, open_obj)
            self.assertEqual(raw_message, json.dumps(notification).encode())
        _handler.call_count = 0
        open_received.connect(_handler)

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = True
            response = self.client.post(reverse("event_webhook"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(_handler.call_count, 1)

    def test_handle_click_event(self):
        """
        Test handling a click event request.
        """
        req_mail_obj, req_click_obj, notification = get_mock_click()

        def _handler(sender, mail_obj, click_obj, raw_message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(req_mail_obj, mail_obj)
            self.assertEqual(req_click_obj, click_obj)
            self.assertEqual(raw_message, json.dumps(notification).encode())
        _handler.call_count = 0
        click_received.connect(_handler)

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = True
            response = self.client.post(reverse("event_webhook"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(_handler.call_count, 1)

    @override_settings(AWS_SES_INBOUND_HANDLER="global.DummyClass")
    def test_handle_received_event(self):
        """
        Test handling a received event request.
        """

        req_mail_obj, req_content, req_receipt_obj, notification = get_mock_received_sns()

        class DummyClass(BaseHandler):
            def process(self):
                pass

            def handle(self):
                pass

        with mock.patch("importlib.import_module") as mock_import_module:
            mock_import_module.return_value = mock.MagicMock(DummyClass=DummyClass)

            with mock.patch.object(DummyClass, "handle") as mock_handle:

                # Mock the verification
                with mock.patch.object(ses_utils, "verify_event_message") as verify:
                    verify.return_value = True
                    response = self.client.post(reverse("event_webhook"), json.dumps(notification),
                                                content_type="application/json")
                self.assertEqual(response.status_code, 200)

                mock_handle.assert_called_once()

    def test_bad_json(self):
        """
        Test request with invalid JSON.
        """
        response = self.client.post(reverse("event_webhook"), "This is not JSON", content_type="application/json")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.content, "The request body could not be deserialized. Bad JSON.".encode())

    def test_bad_signature(self):
        """
        Test handling a click event request.
        """
        req_mail_obj, req_click_obj, notification = get_mock_click()

        # Mock the verification
        with mock.patch.object(ses_utils, "verify_event_message") as verify:
            verify.return_value = False
            response = self.client.post(reverse("event_webhook"), json.dumps(notification),
                                        content_type="application/json")
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.content, "Signature verification failed.".encode())



from django.core.mail import send_mail
from django.test import TestCase, override_settings

from django_ses import models, signals
from tests.mocks import get_mock_bounce, get_mock_complaint


class SignalsTestCase(TestCase):
    """
    Test the signals handlers.
    """
    def test_bounce_handler(self):
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 0)

        mail_obj, bounce_obj, notification = get_mock_bounce("eventType")

        # It shouldn't insert the email in the blacklist if the feature hasn't
        # been enabled
        signals.bounce_handler(None, mail_obj, bounce_obj, notification)
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 0)

        # ... but after enabling the feature...
        with override_settings(AWS_SES_ADD_BOUNCE_TO_BLACKLIST=True):
            signals.bounce_handler(None, mail_obj, bounce_obj, notification)

        # ... it should insert the email in the blacklist
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 1)

    @override_settings(AWS_SES_ADD_BOUNCE_TO_BLACKLIST=True)
    def test_bounce_handler_twice(self):
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 0)

        mail_obj, bounce_obj, notification = get_mock_bounce("eventType")

        # ... it should insert the email in the blacklist
        signals.bounce_handler(None, mail_obj, bounce_obj, notification)
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 1)

        # ... it shouldn't insert the email in the blacklist again
        signals.bounce_handler(None, mail_obj, bounce_obj, notification)
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 1)

    def test_complaint_handler(self):
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 0)

        mail_obj, complaint_obj, notification = get_mock_complaint("eventType")

        # It shouldn't insert the email in the blacklist if the feature hasn't
        # been enabled
        signals.complaint_handler(None, mail_obj, complaint_obj, notification)
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 0)

        # ... but after enabling the feature...
        with override_settings(AWS_SES_ADD_COMPLAINT_TO_BLACKLIST=True):
            signals.complaint_handler(None, mail_obj, complaint_obj, notification)

        # ... it should insert the email in the blacklist
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 1)

    @override_settings(AWS_SES_ADD_COMPLAINT_TO_BLACKLIST=True)
    def test_complaint_handler_twice(self):
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 0)

        mail_obj, complaint_obj, notification = get_mock_complaint("eventType")

        # ... it should insert the email in the blacklist
        signals.complaint_handler(None, mail_obj, complaint_obj, notification)
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 1)

        # ... it shouldn't insert the email in the blacklist again
        signals.complaint_handler(None, mail_obj, complaint_obj, notification)
        count = models.BlacklistedEmail.objects.all().count()
        self.assertEqual(count, 1)

    @override_settings(EMAIL_BACKEND='tests.test_backend.FakeSESBackend')
    def test_message_sent_handler(self):
        def _handler(sender, message, **kwargs):
            _handler.call_count += 1
            self.assertEqual(message.subject, 'subject')
            self.assertEqual(message.body, 'body')
            self.assertEqual(message.extra_headers['message_id'], 'fake_message_id')

        _handler.call_count = 0
        signals.message_sent.connect(_handler)

        send_mail('subject', 'body', 'from@example.com', ['to@example.com'])
        self.assertEqual(_handler.call_count, 1)



from django.apps import AppConfig


class DjangoSESConfig(AppConfig):
    name = 'django_ses'
    verbose_name = 'Django SES'

    default_auto_field = 'django.db.models.AutoField'

    def ready(self):
        # Explicitly connect signal handlers decorated with @receiver.
        from django_ses.signals import (
            bounce_handler,
            bounce_received,
            complaint_handler,
            complaint_received,
        )
        bounce_received.connect(bounce_handler)
        complaint_received.connect(complaint_handler)



import datetime
from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from django_ses.management.commands import get_ses_statistics as mod_get_ses_statistics
from django_ses.models import BlacklistedEmail, SESStat

data_points = [
    {
        'Complaints': '1',
        'Timestamp': datetime.datetime(year=2012, month=1, day=1, hour=2),
        'DeliveryAttempts': '2',
        'Bounces': '3',
        'Rejects': '4'
    },
    {
        'Complaints': '1',
        'Timestamp': datetime.datetime(year=2012, month=1, day=3, hour=2),
        'DeliveryAttempts': '2',
        'Bounces': '3',
        'Rejects': '4'
    },
    {
        'Complaints': '1',
        'Timestamp': datetime.datetime(year=2012, month=1, day=3, hour=3),
        'DeliveryAttempts': '2',
        'Bounces': '3',
        'Rejects': '4'
    }
]


class FakeSESConnection:
    def __init__(self, *args, **kwargs):
        pass

    def get_send_statistics(self):
        return {
            'SendDataPoints': data_points,
            'ResponseMetadata': {
                'RequestId': '1'
            },
        }


class SESCommandTest(TestCase):

    def setUp(self):
        mod_get_ses_statistics.boto3.client = FakeSESConnection

    def test_get_statistics(self):
        # Test the get_ses_statistics management command
        call_command('get_ses_statistics')

        # Test that days with a single data point is saved properly
        stat = SESStat.objects.get(date='2012-01-01')
        self.assertEqual(stat.complaints, 1)
        self.assertEqual(stat.delivery_attempts, 2)
        self.assertEqual(stat.bounces, 3)
        self.assertEqual(stat.rejects, 4)

        # Test that days with multiple data points get saved properly
        stat = SESStat.objects.get(date='2012-01-03')
        self.assertEqual(stat.complaints, 2)
        self.assertEqual(stat.delivery_attempts, 4)
        self.assertEqual(stat.bounces, 6)
        self.assertEqual(stat.rejects, 8)

        # Changing data points should update database records too
        data_points[0]['Complaints'] = '2'
        data_points[0]['DeliveryAttempts'] = '3'
        data_points[0]['Bounces'] = '4'
        data_points[0]['Rejects'] = '5'

        call_command('get_ses_statistics')
        stat = SESStat.objects.get(date='2012-01-01')
        self.assertEqual(stat.complaints, 2)
        self.assertEqual(stat.delivery_attempts, 3)
        self.assertEqual(stat.bounces, 4)
        self.assertEqual(stat.rejects, 5)


class BlacklistCommandRTest(TestCase):
    def test_add_command(self):
        blacklisted = BlacklistedEmail.objects.all().count()
        self.assertEqual(blacklisted, 0)

        call_command('blacklist', '--add', 'foo@bar.com')

        blacklisted = BlacklistedEmail.objects.get(email='foo@bar.com')
        self.assertEqual(blacklisted.email, 'foo@bar.com')

    def test_delete_command(self):
        BlacklistedEmail.objects.create(email='foo@bar.com')
        BlacklistedEmail.objects.create(email='foo2@bar.com')
        BlacklistedEmail.objects.create(email='foo3@bar.com')

        blacklisted = BlacklistedEmail.objects.get(email='foo@bar.com')
        self.assertEqual(blacklisted.email, 'foo@bar.com')

        call_command('blacklist', '--delete', 'foo@bar.com')

        blacklisted = BlacklistedEmail.objects.filter(email='foo@bar.com').count()
        self.assertEqual(blacklisted, 0)

        blacklisted = BlacklistedEmail.objects.all().count()
        self.assertEqual(blacklisted, 2)

    def test_list_command(self):
        BlacklistedEmail.objects.create(email='foo@bar.com')
        BlacklistedEmail.objects.create(email='foo2@bar.com')
        BlacklistedEmail.objects.create(email='foo3@bar.com')

        out = StringIO()
        call_command('blacklist', '--list', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 3)
        self.assertIn('foo@bar.com', lines)
        self.assertIn('foo2@bar.com', lines)
        self.assertIn('foo3@bar.com', lines)

    def test_list_command_pagination(self):
        for i in range(55):
            BlacklistedEmail.objects.create(email=f'foo{i}@bar.com')

        out = StringIO()
        call_command('blacklist', '--list', '--page', '1', '--limit', '10', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 10)
        for i in range(10):
            self.assertIn(f'foo{i}@bar.com', lines)

        out = StringIO()
        call_command('blacklist', '--list', '--page', '2', '--limit', '10', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 10)
        for i in range(10, 20):
            self.assertIn(f'foo{i}@bar.com', lines)

        out = StringIO()
        call_command('blacklist', '--list', '--page', '6', '--limit', '10', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 5)
        for i in range(50, 55):
            self.assertIn(f'foo{i}@bar.com', lines)

        out = StringIO()
        call_command('blacklist', '--list', '--limit', '0', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 55)
        for i in range(55):
            self.assertIn(f'foo{i}@bar.com', lines)

    def test_search_command(self):
        BlacklistedEmail.objects.create(email='aaa@foo.com')
        BlacklistedEmail.objects.create(email='aaa@bar.com')
        BlacklistedEmail.objects.create(email='bbb@foo.com')
        BlacklistedEmail.objects.create(email='bbb@bar.com')
        BlacklistedEmail.objects.create(email='ccc@foo.com')
        BlacklistedEmail.objects.create(email='ccc@bar.com')

        out = StringIO()
        call_command('blacklist', '--search', 'bbb', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 2)
        self.assertIn('bbb@foo.com', lines)
        self.assertIn('bbb@bar.com', lines)

    def test_search_command_pagination(self):
        for i in range(55):
            BlacklistedEmail.objects.create(email=f'foo{i}@bar.com')

        out = StringIO()
        call_command('blacklist', '--search', 'foo', '--page', '1', '--limit', '10', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 10)
        for i in range(10):
            self.assertIn(f'foo{i}@bar.com', lines)

        out = StringIO()
        call_command('blacklist', '--search', 'foo', '--page', '2', '--limit', '10', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 10)
        for i in range(10, 20):
            self.assertIn(f'foo{i}@bar.com', lines)

        out = StringIO()
        call_command('blacklist', '--search', 'foo', '--page', '6', '--limit', '10', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 5)
        for i in range(50, 55):
            self.assertIn(f'foo{i}@bar.com', lines)

        out = StringIO()
        call_command('blacklist', '--search', 'foo', '--limit', '0', stdout=out)
        lines = out.getvalue().strip().splitlines()[1:]
        self.assertEqual(len(lines), 55)
        for i in range(55):
            self.assertIn(f'foo{i}@bar.com', lines)



from django.test import TestCase

from django_ses.inbound import S3Handler, SnsHandler, UnprocessableError
from tests.mocks import (
    get_mock_received_s3,
    get_mock_received_sns,
)


class InbounceSnsTestCase(TestCase):
    def test_happy_path(self):
        mail_obj, content, receipt, notification = get_mock_received_sns()

        class MyReceiver(SnsHandler):
            def process(self):
                self.test_result_subject = self.email.get("subject")
                self.test_result_plain_text = self.email.get("plain_text")
                self.test_result_attachments = self.email.get("attachments")

        handler = MyReceiver(
            mail_obj=mail_obj, receipt=receipt, raw_message=notification)
        handler.handle(content=content)

        self.assertEqual(handler.test_result_subject, "test")
        self.assertEqual(handler.test_result_plain_text, "hey!!\n")
        self.assertEqual(len(handler.test_result_attachments), 1)
        attachment = handler.test_result_attachments[0]
        self.assertEqual(attachment.get("filename"), "attachmnet.png")
        self.assertEqual(attachment.get("content_type"), "image/png")
        self.assertEqual(len(attachment.get("data")), 3013)

    def test_bad_content(self):
        mail_obj, content, receipt, notification = get_mock_received_sns()
        handler = SnsHandler(
            mail_obj=mail_obj, receipt=receipt, raw_message=notification)

        with self.assertLogs('django_ses', level='INFO') as cm:
          with self.assertRaises(UnprocessableError):
              handler.handle(content="bad content that is not base64!")

          self.assertEqual(
              cm.output[-1],
              "ERROR:django_ses.inbound:Content couldn't be decoded with base64. Invalid content?")

    def test_wrong_handler(self):
        mail_obj, content, receipt, notification = get_mock_received_s3()
        handler = SnsHandler(
            mail_obj=mail_obj, receipt=receipt, raw_message=notification)

        with self.assertLogs('django_ses', level='INFO') as cm:
          with self.assertRaises(UnprocessableError):
              handler.handle(content=content)

          self.assertEqual(
              cm.output[-1],
              "ERROR:django_ses.inbound:Received action type (S3) or encoding (None) can't be handled by this handler")


class InbounceS3TestCase(TestCase):
    def test_happy_path(self):
        mail_obj, content, receipt, notification = get_mock_received_s3()

        class MyReceiver(S3Handler):
            def prepare_content(self, _):
                return b"hey!!\n"

            def process(self):
                self.test_result_subject = self.email.get("subject")
                self.test_result_plain_text = self.email.get("plain_text")
                self.test_result_attachments = self.email.get("attachments")

        handler = MyReceiver(
            mail_obj=mail_obj, receipt=receipt, raw_message=notification)
        handler.handle(content=content)

        self.assertEqual(handler.test_result_subject, "test")
        self.assertEqual(handler.test_result_plain_text, "hey!!\n")
        self.assertEqual(handler.test_result_attachments, [])

    def test_wrong_handler(self):
        mail_obj, content, receipt, notification = get_mock_received_sns()
        handler = S3Handler(
            mail_obj=mail_obj, receipt=receipt, raw_message=notification)

        with self.assertLogs('django_ses', level='INFO') as cm:
          with self.assertRaises(UnprocessableError):
              handler.handle(content=content)

          self.assertEqual(
              cm.output[-1],
              "ERROR:django_ses.inbound:Received action type (SNS) can't be handled by this handler")



from django.test import TestCase, override_settings

from django_ses import settings


class SettingsImportTest(TestCase):
    @override_settings(
        AWS_ACCESS_KEY_ID='<<access_key>>',
        AWS_SECRET_ACCESS_KEY='<<secret_key>>',
    )
    def test_aws_access_key_given(self):
        self.assertEqual(settings.ACCESS_KEY, '<<access_key>>')
        self.assertEqual(settings.SECRET_KEY, '<<secret_key>>')

    @override_settings(
        AWS_SES_ACCESS_KEY_ID='<<access_key>>',
        AWS_SES_SECRET_ACCESS_KEY='<<secret_key>>',
    )
    def test_ses_access_key_given(self):
        self.assertEqual(settings.ACCESS_KEY, '<<access_key>>')
        self.assertEqual(settings.SECRET_KEY, '<<secret_key>>')

    @override_settings(AWS_SESSION_TOKEN='<<session_token>>')
    def test_aws_session_token_given(self):
        self.assertEqual(settings.SESSION_TOKEN, '<<session_token>>')

    @override_settings(AWS_SES_SESSION_TOKEN='<<session_token>>')
    def test_ses_session_token_given(self):
        self.assertEqual(settings.SESSION_TOKEN, '<<session_token>>')

    @override_settings(AWS_SES_CONFIGURATION_SET='<<config_set>>')
    def test_ses_configuration_set_given(self):
        self.assertEqual(settings.AWS_SES_CONFIGURATION_SET, '<<config_set>>')

    def test_ses_region_to_endpoint_default_given(self):
        self.assertEqual(settings.AWS_SES_REGION_NAME, 'us-east-1')
        self.assertEqual(
            settings.AWS_SES_REGION_ENDPOINT,
            f'email.{settings.AWS_SES_REGION_NAME}.amazonaws.com',
        )

    @override_settings(AWS_SES_REGION_NAME='eu-west-1')
    def test_ses_region_to_endpoint_set_given(self):
        self.assertEqual(settings.AWS_SES_REGION_NAME, 'eu-west-1')
        self.assertEqual(settings.AWS_SES_REGION_ENDPOINT, 'email.eu-west-1.amazonaws.com')



from django.db import NotSupportedError
from django.dispatch import Signal

from django_ses.conf import settings

# The following fields are used from the 3 signals below: mail_obj, bounce_obj, raw_message
message_sent = Signal()
bounce_received = Signal()
complaint_received = Signal()
delivery_received = Signal()
send_received = Signal()
open_received = Signal()
click_received = Signal()


def _blacklist_recipients(recipients):
    from django_ses import models

    if len(recipients) == 0:
        return

    recipients = [email.lower() for email in recipients]

    # If we received only 1 recipients, don't waste extra queries trying to find
    # out if the recipient has already been blacklisted; just attempt to
    # blacklist it.
    if len(recipients) == 1:
        models.BlacklistedEmail.objects.get_or_create(email=recipients[0])
        return

    qs = models.BlacklistedEmail.objects.filter(email__in=recipients)
    blacklisted_emails = set(qs.values_list("email", flat=True))
    unblacklisted_emails = [email for email in recipients if email not in blacklisted_emails]

    # Try to bulk-insert the unblacklisted emails. If the operation fails (a
    # possibility only when using Oracle because it doesn't support
    # ignore_conflicts), fallback to one-by-one insertion.
    try:
        models.BlacklistedEmail.objects.bulk_create([
            models.BlacklistedEmail(email=email) for email in unblacklisted_emails
        ], ignore_conflicts=True)
    except NotSupportedError: # Oracle doesn't support "ignore_conflicts"
        for email in recipients:
            models.BlacklistedEmail.objects.get_or_create(email=email)


def bounce_handler(sender, mail_obj, bounce_obj, raw_message, *args, **kwargs):
    if not settings.AWS_SES_ADD_BOUNCE_TO_BLACKLIST:
        return

    from django_ses.utils import get_permanent_bounced_emails_from_bounce_obj
    bounced_recipients = get_permanent_bounced_emails_from_bounce_obj(bounce_obj)
    _blacklist_recipients(bounced_recipients)


def complaint_handler(sender, mail_obj, complaint_obj, raw_message, *args, **kwargs):
    if not settings.AWS_SES_ADD_COMPLAINT_TO_BLACKLIST:
        return

    from django_ses.utils import get_emails_from_complaint_obj
    complaint_emails = get_emails_from_complaint_obj(complaint_obj)
    _blacklist_recipients(complaint_emails)



import base64
import logging
import re
import warnings
from builtins import bytes
from email.utils import parseaddr
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import urlopen

from django.core.exceptions import ImproperlyConfigured

from django_ses import settings
from django_ses.deprecation import RemovedInDjangoSES20Warning

logger = logging.getLogger(__name__)

_CERT_CACHE = {}

SES_REGEX_CERT_URL = re.compile(
    r"(?i)^https://sns\.[a-z0-9\-]+\.amazonaws\.com(\.cn)?/SimpleNotificationService\-[a-z0-9]+\.pem$"
)


def clear_cert_cache():
    """Clear the certificate cache.

    This one-liner exists to discourage imports and direct usage of
    _CERT_CACHE.

    :returns None
    """
    _CERT_CACHE.clear()


class EventMessageVerifier(object):
    """
    A utility class for validating event messages

    See: http://docs.amazonwebservices.com/sns/latest/gsg/SendMessageToHttp.verify.signature.html
    """

    _REQ_DEP_TMPL = (
        "%s is required for event message verification. Please install "
        "`django-ses` with the `event` extra - e.g. "
        "`pip install django-ses[events]`."
    )

    def __init__(self, notification):
        """
        Creates a new event message from the given dict.
        """
        self._data = notification
        self._verified = None

    def is_verified(self):
        """
        Verifies an SES event message.

        Sign the bytes from the notification and compare it to the signature in
        the notification. If same, return True; else False.
        """
        if self._verified is not None:
            return self._verified

        signature = self._data.get("Signature")
        if not signature:
            self._verified = False
            return self._verified

        # Decode the signature from base64
        signature = bytes(base64.b64decode(signature))

        # Get the message to sign
        sign_bytes = self._get_bytes_to_sign()
        if not sign_bytes:
            self._verified = False
            return self._verified

        if not self.certificate:
            self._verified = False
            return self._verified

        try:
            from cryptography.exceptions import InvalidSignature
            from cryptography.hazmat.primitives import hashes
            from cryptography.hazmat.primitives.asymmetric import padding
        except ImportError:
            raise ImproperlyConfigured(self._REQ_DEP_TMPL % "`cryptography`")

        # Extract the public key
        pkey = self.certificate.public_key()

        # Use the public key to verify the signature.
        try:
            # The details here do not appear to be documented, but the
            # algorithm and padding choices work in testing, which should mean
            # they're the right ones.
            pkey.verify(
                signature,
                sign_bytes,
                padding.PKCS1v15(),
                hashes.SHA1(),
            )
        except InvalidSignature:
            logger.warning(
                "Invalid signature on message with ID: %s",
                self._data.get("MessageId"),
            )
            self._verified = False
        else:
            self._verified = True
        return self._verified

    @property
    def certificate(self):
        """
        Retrieves the certificate used to sign the event message.

        :returns: None if the cert cannot be retrieved. Else, gets the cert
        caches it, and returns it, or simply returns it if already cached.
        """
        cert_url = self._get_cert_url()
        if not cert_url:
            return None

        if cert_url in _CERT_CACHE:
            return _CERT_CACHE[cert_url]

        # Only load certificates from a certain domain?
        # Without some kind of trusted domain check, any old joe could
        # craft a event message and sign it using his own certificate
        # and we would happily load and verify it.
        try:
            import requests
            from requests import RequestException
        except ImportError:
            raise ImproperlyConfigured(self._REQ_DEP_TMPL % "`requests`")

        try:
            from cryptography import x509
        except ImportError:
            raise ImproperlyConfigured(self._REQ_DEP_TMPL % "`cryptography`")

        # We use requests because it verifies the https certificate when
        # retrieving the signing certificate. If https was somehow hijacked
        # then all bets are off.
        try:
            response = requests.get(cert_url, timeout=10)
            response.raise_for_status()
        except RequestException as exc:
            logger.warning(
                "Network error downloading certificate from " "%s: %s",
                cert_url,
                exc,
            )
            _CERT_CACHE[cert_url] = None
            return _CERT_CACHE[cert_url]

        # Handle errors loading the certificate.
        # If the certificate is invalid then return
        # false as we couldn't verify the message.
        try:
            _CERT_CACHE[cert_url] = x509.load_pem_x509_certificate(response.content)
        except ValueError as e:
            logger.warning('Could not load certificate from %s: "%s"', cert_url, e)
            _CERT_CACHE[cert_url] = None

        return _CERT_CACHE[cert_url]

    def _get_cert_url(self):
        """
        Get the signing certificate URL.
        Only accept urls that match the domains set in the
        AWS_SNS_EVENT_CERT_TRUSTED_DOMAINS setting. Sub-domains
        are allowed. i.e. if amazonaws.com is in the trusted domains
        then sns.us-east-1.amazonaws.com will match.
        """
        cert_url = self._data.get("SigningCertURL")
        if not cert_url:
            logger.warning('No signing certificate URL: "%s"', cert_url)
            return None

        if not cert_url.startswith("https://"):
            logger.warning('Untrusted certificate URL: "%s"', cert_url)
            return None

        url_obj = urlparse(cert_url)
        for trusted_domain in settings.EVENT_CERT_DOMAINS:
            parts = trusted_domain.split(".")
            if "amazonaws.com" in trusted_domain:
                if not SES_REGEX_CERT_URL.match(cert_url):
                    if len(parts) < 4:
                        return None
                    else:
                        logger.warning('Possible security risk for: "%s"', cert_url)
                        logger.warning(
                            "It is strongly recommended to configure the full domain in EVENT_CERT_DOMAINS. "
                            "See v3.5.0 release notes for more details."
                        )

            if url_obj.netloc.split(".")[-len(parts) :] == parts:
                return cert_url

        return None

    def _get_bytes_to_sign(self):
        """
        Creates the message used for signing SNS notifications.
        This is used to verify the bounce message when it is received.
        """

        # Depending on the message type the fields to add to the message
        # differ so we handle that here.
        msg_type = self._data.get("Type")
        if msg_type == "Notification":
            fields_to_sign = [
                "Message",
                "MessageId",
                "Subject",
                "Timestamp",
                "TopicArn",
                "Type",
            ]
        elif (
            msg_type == "SubscriptionConfirmation"
            or msg_type == "UnsubscribeConfirmation"
        ):
            fields_to_sign = [
                "Message",
                "MessageId",
                "SubscribeURL",
                "Timestamp",
                "Token",
                "TopicArn",
                "Type",
            ]
        else:
            # Unrecognized type
            logger.warning('Unrecognized SNS message Type: "%s"', msg_type)
            return None

        bytes_to_sign = []
        for field in fields_to_sign:
            field_value = self._data.get(field)
            if not field_value:
                continue

            # Some notification types do not have all fields. Only add fields
            # with values.
            bytes_to_sign.append(f"{field}\n{field_value}\n")

        return "".join(bytes_to_sign).encode()


def BounceMessageVerifier(*args, **kwargs):  # noqa: N802
    warnings.warn(
        "utils.BounceMessageVerifier is deprecated. It is renamed to EventMessageVerifier.",
        RemovedInDjangoSES20Warning,
    )

    # parameter name is renamed from bounce_dict to notification.
    if "bounce_dict" in kwargs:
        kwargs["notification"] = kwargs["bounce_dict"]
        del kwargs["bounce_dict"]

    return EventMessageVerifier(*args, **kwargs)


def verify_event_message(notification):
    """
    Verify an SES/SNS event notification message.
    """
    verifier = EventMessageVerifier(notification)
    return verifier.is_verified()


def verify_bounce_message(msg):
    """
    Verify an SES/SNS bounce(event) notification message.
    """
    warnings.warn(
        "utils.verify_bounce_message is deprecated. It is renamed to verify_event_message.",
        RemovedInDjangoSES20Warning,
    )
    return verify_event_message(msg)


def confirm_sns_subscription(notification):
    logger.info(
        "Received subscription confirmation: TopicArn: %s",
        notification.get("TopicArn"),
        extra={
            "notification": notification,
        },
    )

    # Get the subscribe url and hit the url to confirm the subscription.
    subscribe_url = notification.get("SubscribeURL")
    try:
        urlopen(subscribe_url).read()
    except URLError as e:
        # Some kind of error occurred when confirming the request.
        logger.error(
            'Could not confirm subscription: "%s"',
            e,
            extra={
                "notification": notification,
            },
            exc_info=True,
        )


def get_permanent_bounced_emails_from_bounce_obj(bounce_obj: dict) -> list:
    """Extracts permanent bounced email addresses only as a list of strings.
    https://docs.aws.amazon.com/ses/latest/DeveloperGuide/notification-contents.html#bounce-object
    """
    bounced_recipients = bounce_obj.get("bouncedRecipients", list())
    return [br.get("emailAddress") for br in bounced_recipients if br.get("status").startswith("5")]


def get_emails_from_complaint_obj(complaint_obj: dict) -> list:
    """Extracts complaint email addresses from complaint_obj
    https://docs.aws.amazon.com/ses/latest/DeveloperGuide/notification-contents.html#complaint-object
    """
    return [cr.get("emailAddress") for cr in complaint_obj.get("complainedRecipients", list())]


def filter_blacklisted_recipients(addresses):
    from django_ses import models

    """Remove blacklisted emails from addresses"""
    if isinstance(addresses, str):
        addresses = [addresses]

    emails = [parseaddr(recipient)[1].lower() for recipient in addresses]
    qs = models.BlacklistedEmail.objects.filter(email__in=emails)
    excluded_emails = set(qs.values_list("email", flat=True))
    return [email for email in emails if email not in excluded_emails]



import json
import os


def get_mock_email():
    return {
        "timestamp": "2012-05-25T14:59:38.623-07:00",
        "messageId": "000001378603177f-7a5433e7-8edb-42ae-af10-f0181f34d6ee-000000",
        "source": "sender@example.com",
        "destination": [
            "recipient1@example.com",
            "recipient2@example.com",
            "recipient3@example.com",
            "recipient4@example.com"
        ]
    }

def get_mock_received_email():
    return {
        'timestamp': '2024-10-27T17:13:33.684Z',
        'source': 'alexandernst@gmail.com',
        'messageId': '0bj0fp4ev5vrddnpflq7a17rpgmo7gnhslacdu01',
        'destination': ['foo@bar.com'],
        'headersTruncated': False,
        'headers': [
            {
                'name': 'Return-Path',
                'value': '<alexandernst@gmail.com>'
            },
            {
                'name': 'Received',
                'value': 'from mail-lf1-f45.google.com '
                         '(mail-lf1-f45.google.com [209.85.167.45]) by '
                         'inbound-smtp.us-east-1.amazonaws.com with '
                         'SMTP id '
                         '0bj0fp4ev5vrddnpflq7a17rpgmo7gnhslacdu01 for '
                         'foo@bar.com; Sun, 27 Oct 2024 '
                         '17:13:33 +0000 (UTC)'
            },
            {
                'name': 'X-SES-Spam-Verdict',
                'value': 'PASS'
            },
            {
                'name': 'X-SES-Virus-Verdict',
                'value': 'PASS'
            },
            {
                'name': 'Received-SPF',
                'value': 'pass (spfCheck: domain of _spf.google.com '
                         'designates 209.85.167.45 as permitted sender) '
                         'client-ip=209.85.167.45; '
                         'envelope-from=alexandernst@gmail.com; '
                         'helo=mail-lf1-f45.google.com;'
            },
            {
                'name': 'Authentication-Results',
                'value': 'amazonses.com; spf=pass (spfCheck: domain of '
                         '_spf.google.com designates 209.85.167.45 as '
                         'permitted sender) client-ip=209.85.167.45; '
                         'envelope-from=alexandernst@gmail.com; '
                         'helo=mail-lf1-f45.google.com; dkim=pass '
                         'header.i=@gmail.com; dmarc=pass '
                         'header.from=gmail.com;'
            },
            {
                'name': 'X-SES-RECEIPT',
                'value': 'AEFBQUFBQUFBQUFGOEpra2RyZWJPUlZ4d3NkTmhyZ2NBM0tlRktEeUNvVmt0NGdiMzNWUk1zWXlUZ2JwVktJU3NDdGZ4c'
                         'kRTMURZc0k3MlBaTlkyamRiY3daRFlxbEtCakFLeGppSGJrQkh0ejJ0N0xBcXJ2TDJlUTRJRHU4anJTRU1wMmpQVTFJZE'
                         'NVWHFiT2NpaVdkSWsxK1FMQTM0MXE2WXkrWW5CN3RHRkJkTDI3eUMvVnRrOWh5MVZRMlBiK2REWWFhY0xYYmVkL3VHYUd'
                         'oYWEvUmVFTytCMHpXUWR1eTJPUjFhQ3BBSHlBL2dieTBwSCt0QTFFRTh2SDVZcVJ1UjY4UDhac0tPZ01UTlRWZ2t2V0py'
                         'VFJCNlB6ZE9XZk10Y1lieFptbFdQV3Jjbk9pdWhqMlE9PQ=='
            },
            {
                'name': 'X-SES-DKIM-SIGNATURE',
                'value': 'a=rsa-sha256; q=dns/txt; '
                         'b=Y7IWMTNTKesfKhopRg2sjUOlVs5NeQKjSGavxaywj3KlWg+6ovOueCYZZK8/4QOmb6X2/sS29JYhrmyUvVXueoV1/iw'
                         'TPlxTDaKdMjsTILCMI+nEvVeV1e+r4t9BgV5EpT/kue2rDCd5eIJNFVB77qOAgM5UDcpt+C+APHuEjb4=; '
                         'c=relaxed/simple; '
                         's=ug7nbtf4gccmlpwj322ax3p6ow6yfsug; '
                         'd=amazonses.com; t=1730049214; v=1; '
                         'bh=3+TxY9fvf0zPnsSER7J7drTP5kKXNEg+Ue1NKN3mII4=; '
                         'h=From:To:Cc:Bcc:Subject:Date:Message-ID:MIME-Version:Content-Type:X-SES-RECEIPT;'
            },
            {
                'name': 'Received',
                'value': 'by mail-lf1-f45.google.com with SMTP id '
                         '2adb3069b0e04-539f84907caso3771922e87.3 for '
                         '<foo@bar.com>; Sun, 27 Oct 2024 '
                         '10:13:33 -0700 (PDT)'
            },
            {
                'name': 'DKIM-Signature',
                'value': 'v=1; a=rsa-sha256; c=relaxed/relaxed; '
                         'd=gmail.com; s=20230601; t=1730049212; '
                         'x=1730654012; darn=bar.com; '
                         'h=to:subject:message-id:date:from:mime-version:from:to:cc:subject:date:message-id:reply-to; '
                         'bh=3+TxY9fvf0zPnsSER7J7drTP5kKXNEg+Ue1NKN3mII4=; '
                         'b=FpIcePmBAbNZQErjJsePpOvUeNXuiimPXy+udQg47WrX3V8YtiKecCOuIjCoxk7AQ9yJYWOkZO4h2cjLGw4eyyMiHUl'
                         'qJC7zO7Uph7SDhPvPptxKXyv3L0Y4sMs/KWg3cHVDqOFX3zfFclsVpocfltJFj2nDDr5gQeXONsAn0nGPRK7b+cTlI2w/'
                         'RJy/q+MNThsjrB4JQurHbGhTcHb+B3cZtkHfClX+PzGZ+DkSsx8pdDZ/SsovGJIKXOgeaagJfO9pdt2muzJBBHBCw0sJ8'
                         '4FGmMbOYWY0OuTTZRugq2OMe51UlAqZOim2bVVWwPAmrD4bQ5rsFK9ufctPsblVDA=='
            },
            {
                'name': 'X-Google-DKIM-Signature',
                'value': 'v=1; a=rsa-sha256; c=relaxed/relaxed; '
                         'd=1e100.net; s=20230601; t=1730049212; '
                         'x=1730654012; '
                         'h=to:subject:message-id:date:from:mime-version:x-gm-message-state '
                         ':from:to:cc:subject:date:message-id:reply-to; '
                         'bh=3+TxY9fvf0zPnsSER7J7drTP5kKXNEg+Ue1NKN3mII4=; '
                         'b=viYdJJci/NjMkwze8dOZmUn5Zerb6LFFUdu6oQUhaOAD2yvtPxjoh4ToKgcomJVjPA '
                         '8Gxk53YQyAvPsYIsTGCBbiU6PzXcfG4cvWqgJw6CdGRJ0POiSXxmMxcljaoISrzSu8Lz '
                         'HHMxQ57FVxq4sud5aBSqIM9Axt3M0nIrbHDc+eOcYHsVd1svc1n9Mbqnftvbj6veLi+W '
                         'yb+GSRJyf4m2zpyeM8d2cLxFj4+IDEtYTmYX6hd8qtqf5uZYoAKfMRMIASY7binH/DRo '
                         'cxsgMi/XVGRxIACi0BFIQqs/pP0E04grULFnfFRE5S1uBv6iMtcQ2H21seQuboyFKqbs '
                         'lIQw=='
            },
            {
                'name': 'X-Gm-Message-State',
                'value': 'AOJu0YyeijzszzCRTvcUvLjjNwVZiIjxnL4vC7NQoMNHxqeCXbMFfYQ6 '
                         'eWFopGZVYLIoc62feiIAC4lU730b5FqR1gUnNo+m0quipOQUTJ/VbnQ52b35fSlv2yysZpqrGPo '
                         '2nxQAeb7zHSKQJodJlcC5zpCH0DhvfQ=='
            },
            {
                'name': 'X-Google-Smtp-Source',
                'value': 'AGHT+IHMZZbLZB56p/cTh66TxY/jz9cltZNQvoRAcbzc5RjUQmYKo8nMnl1LLcUlZcx0HsntJ8vAf2CbXPfqoT5rpPU='
            },
            {
                'name': 'X-Received',
                'value': 'by 2002:a05:6512:238a:b0:539:f949:c027 with '
                         'SMTP id '
                         '2adb3069b0e04-53b348f9b74mr2135393e87.18.1730049211812; '
                         'Sun, 27 Oct 2024 10:13:31 -0700 (PDT)'
            },
            {
                'name': 'MIME-Version',
                'value': '1.0'
            },
            {
                'name': 'From',
                'value': '"Alexander Nestorov (alexandernst)" <alexandernst@gmail.com>'
            },
            {
                'name': 'Date',
                'value': 'Sun, 27 Oct 2024 18:12:55 +0100'
            },
            {
                'name': 'Message-ID',
                'value': '<CACuz9s0EbFBFEdgJN6Pfc74mQ6+-S36UEZXO1=pbabnGx5ObOw@mail.gmail.com>'
            },
            {
                'name': 'Subject',
                'value': 'test'
            },
            {
                'name': 'To',
                'value': 'foo@bar.com'
            },
            {
                'name': 'Content-Type',
                'value': 'multipart/alternative; boundary="000000000000b6d36a0625787884"'
            }
        ],
        'commonHeaders': {
            'returnPath': 'alexandernst@gmail.com',
            'from': [
                '"Alexander Nestorov (alexandernst)" <alexandernst@gmail.com>'
            ],
            'date': 'Sun, 27 Oct 2024 18:12:55 +0100',
            'to': [
                'foo@bar.com'
            ],
            'messageId': '<CACuz9s0EbFBFEdgJN6Pfc74mQ6+-S36UEZXO1=pbabnGx5ObOw@mail.gmail.com>',
            'subject': 'test'
        }
    }


def get_mock_receipt_s3():
    return {
        "timestamp": "2012-05-25T14:59:38.605-07:00",
        "processingTimeMillis": 825,
        "recipients": ["foo@bar.com"],
        "spamVerdict": {"status": "PASS"},
        "virusVerdict": {"status": "PASS"},
        "spfVerdict": {"status": "PASS"},
        "dkimVerdict": {"status": "PASS"},
        "dmarcVerdict": {"status": "PASS"},
        "action": {
            "type": "S3",
            "topicArn": "arn:aws:sns:us-east-1:112310073099:inbox-foo-bar",
            "bucketName": "django-ses-test-inbox",
            "objectKey": "0bj0fp4ev5vrddnpflq7a17rpgmo7gnhslacdu01"
        }
    }


def get_mock_receipt_sns():
    return {
        "timestamp": "2012-05-25T14:59:38.605-07:00",
        "processingTimeMillis": 825,
        "recipients": ["foo@bar.com"],
        "spamVerdict": {"status": "PASS"},
        "virusVerdict": {"status": "PASS"},
        "spfVerdict": {"status": "PASS"},
        "dkimVerdict": {"status": "PASS"},
        "dmarcVerdict": {"status": "PASS"},
        "action": {
            "type": "SNS",
            "topicArn": "arn:aws:sns:us-east-1:112310073099:inbox-foo-bar",
            "encoding": "BASE64",
        }
    }


def get_mock_notification(message):
    return {
        "Type": "Notification",
        "MessageId": "22b80b92-fdea-4c2c-8f9d-bdfb0c7bf324",
        "TopicArn": "arn:aws:sns:us-east-1:123456789012:MyTopic",
        "Subject": "Amazon SES Email Event Notification",
        "Message": json.dumps(message),
        "Timestamp": "2012-05-02T00:54:06.655Z",
        "SignatureVersion": "1",
        "Signature": "",
        "SigningCertURL": "",
        "UnsubscribeURL": ""
    }


def get_mock_bounce(type_specifier):
    mail = get_mock_email()
    bounce = {
        "bounceType": "Permanent",
        "bounceSubType": "General",
        "bouncedRecipients": [
            {
                "status": "5.0.0",
                "action": "failed",
                "diagnosticCode": "smtp; 550 user unknown",
                "emailAddress": "recipient1@example.com",
            },
            {
                "status": "4.0.0",
                "action": "delayed",
                "emailAddress": "recipient2@example.com",
            }
        ],
        "reportingMTA": "example.com",
        "timestamp": "2012-05-25T14:59:38.605-07:00",
        "feedbackId": "000001378603176d-5a4b5ad9-6f30-4198-a8c3-b1eb0c270a1d-000000",
    }

    message = {
        type_specifier: "Bounce",
        "mail": mail,
        "bounce": bounce,
    }
    notification = get_mock_notification(message)
    return mail, bounce, notification

def get_mock_complaint(type_specifier):
    mail = get_mock_email()
    complaint = {
        "userAgent": "ExampleCorp Feedback Loop (V0.01)",
        "complainedRecipients": [
            {
                "emailAddress": "recipient1@example.com"
            }
        ],
        "complaintFeedbackType": "abuse",
        "arrivalDate": "2009-12-03T04:24:21.000-05:00",
        "timestamp": "2012-05-25T14:59:38.623Z",
        "feedbackId": "000001378603177f-18c07c78-fa81-4a58-9dd1-fedc3cb8f49a-000000"
    }

    message = {
        type_specifier: "Complaint",
        "mail": mail,
        "complaint": complaint,
    }
    notification = get_mock_notification(message)
    return mail, complaint, notification


def get_mock_send():
    mail = get_mock_email()
    send = {}
    message = {
        "eventType": "Send",
        "mail": mail,
        "send": send,
    }
    notification = get_mock_notification(message)
    return mail, send, notification


def get_mock_delivery():
    mail = get_mock_email()
    delivery = {
        "timestamp": "2021-01-26T14:38:39.270Z",
        "processingTimeMillis": 1796,
        "recipients": ["recipient1@example.com"],
        "smtpResponse": "",
        "reportingMTA": ""
    }
    message = {
        "eventType": "Delivery",
        "mail": mail,
        "delivery": delivery,
    }
    notification = get_mock_notification(message)
    return mail, delivery, notification


def get_mock_open():
    mail = get_mock_email()
    open = {
        "timestamp": "2021-01-26T14:38:52.115Z",
        "userAgent": "Mozilla/5.0 (Windows NT 5.1; rv:11.0) Gecko Firefox/11.0 (via ggpht.com GoogleImageProxy)",
        "ipAddress": "11.111.11.111"
    }
    message = {
        "eventType": "Open",
        "mail": mail,
        "open": open,
    }
    notification = get_mock_notification(message)
    return mail, open, notification


def get_mock_click():
    mail = get_mock_email()
    click = {
        "timestamp": "2021-01-26T14:38:55.540Z",
        "ipAddress": "11.111.11.111",
        "userAgent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) "
                     "Version/14.0 Safari/605.1.15",
        "link": "github.com/django-ses/django-ses",
        "linkTags": None
    }
    message = {
        "eventType": "Click",
        "mail": mail,
        "click": click,
    }
    notification = get_mock_notification(message)
    return mail, click, notification


def get_mock_received_s3():
    mail = get_mock_received_email()
    receipt = get_mock_receipt_s3()

    message = {
        "notificationType": "Received",
        "mail": mail,
        "receipt": receipt,
    }
    notification = get_mock_notification(message)
    return mail, None, receipt, notification


def get_mock_received_sns():
    mail = get_mock_received_email()
    receipt = get_mock_receipt_sns()

    fpath = os.path.join(os.path.dirname(__file__), "email_content.base64")
    content = open(fpath, "r").read()

    message = {
        "notificationType": "Received",
        "mail": mail,
        "receipt": receipt,
        "content": content,
    }
    notification = get_mock_notification(message)
    return mail, content, receipt, notification



#!/usr/bin/env python
# encoding: utf-8

from django.core.management.base import BaseCommand
from django.core.paginator import Paginator

from django_ses import models


def _add_options(target):
    return (
        target(
            '-a',
            '--add',
            dest='email_to_add',
            default=False,
            help='Adds an email to your blacklist'
        ),
        target(
            '-d',
            '--delete',
            dest='email_to_delete',
            default=False,
            help='Removes an email from your blacklist'
        ),
        target(
            '-l',
            '--list',
            dest='list_emails',
            default=False,
            action='store_true',
            help='Outputs all blacklisted emails'
        ),
        target(
            '-s',
            '--search',
            dest='search_email',
            default=False,
            help='Search for blacklisted emails'
        ),
        target(
            '--page',
            dest='page',
            default=1,
            type=int,
            help='Page in the results. Starts at 1.'
        ),
        target(
            '--limit',
            dest='limit',
            default=1000,
            type=int,
            help='Number of displayed emails per page. Use 0 for unlimited.'
        )
    )


class Command(BaseCommand):
    """Add, delete or list blacklisted email addresses"""

    def add_arguments(self, parser):
        _add_options(parser.add_argument)

    def handle(self, *args, verbosity=0, email_to_add='', email_to_delete='',
               list_emails='', search_email='', page=1,
               limit=1000, **options):
        if email_to_add:
            if verbosity != '0':
                self.stdout.write(f"Adding email: {email_to_add}")
            models.BlacklistedEmail.objects.get_or_create(email=email_to_add.lower())
        elif email_to_delete:
            if verbosity != '0':
                self.stdout.write(f"Removing email {email_to_delete}")
            models.BlacklistedEmail.objects.filter(email=email_to_delete.lower()).delete()
        elif list_emails:
            if verbosity != '0':
                self.stdout.write("Listing blacklisted emails:")

            objs = models.BlacklistedEmail.objects.all().order_by("pk")

            if limit > 0:
                paginator = Paginator(objs, per_page=limit)
                objs = paginator.page(page).object_list

            for obj in objs:
                self.stdout.write(obj.email)
        elif search_email:
            if verbosity != '0':
                self.stdout.write("Searching blacklisted emails:")
            objs = models.BlacklistedEmail.objects.filter(
                email__icontains=search_email).order_by("pk")

            if limit > 0:
                paginator = Paginator(objs, per_page=limit)
                objs = paginator.page(page).object_list

            for obj in objs:
                self.stdout.write(obj.email)



from typing import List, Optional

from django.conf import settings as django_settings

# Allow uppercase variable names for settings properties
# ruff: noqa: N802


class SesSettings:
    @property
    def AWS_SESSION_PROFILE(self) -> Optional[str]:
        return getattr(django_settings, 'AWS_SESSION_PROFILE', None)

    @property
    def ACCESS_KEY(self) -> Optional[str]:
        return getattr(
            django_settings,
            'AWS_SES_ACCESS_KEY_ID',
            getattr(django_settings, 'AWS_ACCESS_KEY_ID', None),
        )

    @property
    def SECRET_KEY(self) -> Optional[str]:
        return getattr(
            django_settings,
            'AWS_SES_SECRET_ACCESS_KEY',
            getattr(django_settings, 'AWS_SECRET_ACCESS_KEY', None),
        )

    @property
    def SESSION_TOKEN(self) -> Optional[str]:
        return getattr(
            django_settings,
            'AWS_SES_SESSION_TOKEN',
            getattr(django_settings, 'AWS_SESSION_TOKEN', None),
        )

    @property
    def AWS_SES_REGION_NAME(self) -> str:
        return getattr(
            django_settings,
            'AWS_SES_REGION_NAME',
            getattr(django_settings, 'AWS_DEFAULT_REGION', 'us-east-1'),
        )

    @property
    def AWS_SES_REGION_ENDPOINT(self) -> str:
        return getattr(django_settings, 'AWS_SES_REGION_ENDPOINT', f'email.{self.AWS_SES_REGION_NAME}.amazonaws.com')

    @property
    def AWS_SES_REGION_ENDPOINT_URL(self) -> str:
        return getattr(django_settings, 'AWS_SES_REGION_ENDPOINT_URL', 'https://' + self.AWS_SES_REGION_ENDPOINT)

    @property
    def AWS_SES_AUTO_THROTTLE(self) -> float:
        return getattr(django_settings, 'AWS_SES_AUTO_THROTTLE', 0.5)

    @property
    def AWS_SES_CONFIG(self):
        return getattr(django_settings, 'AWS_SES_CONFIG', None)

    @property
    def AWS_SES_RETURN_PATH(self) -> Optional[str]:
        return getattr(django_settings, 'AWS_SES_RETURN_PATH', None)

    @property
    def AWS_SES_CONFIGURATION_SET(self) -> Optional[str]:
        return getattr(django_settings, 'AWS_SES_CONFIGURATION_SET', None)

    @property
    def DKIM_DOMAIN(self) -> Optional[str]:
        return getattr(django_settings, 'DKIM_DOMAIN', None)

    @property
    def DKIM_PRIVATE_KEY(self) -> Optional[str]:
        return getattr(django_settings, 'DKIM_PRIVATE_KEY', None)

    @property
    def DKIM_SELECTOR(self) -> str:
        return getattr(django_settings, 'DKIM_SELECTOR', 'ses')

    @property
    def DKIM_HEADERS(self) -> List[str]:
        return getattr(django_settings, 'DKIM_HEADERS', ['From', 'To', 'Cc', 'Subject'])

    @property
    def AWS_SES_SOURCE_ARN(self) -> Optional[str]:
        return getattr(django_settings, 'AWS_SES_SOURCE_ARN', None)

    @property
    def AWS_SES_FROM_ARN(self) -> Optional[str]:
        return getattr(django_settings, 'AWS_SES_FROM_ARN', None)

    @property
    def AWS_SES_RETURN_PATH_ARN(self) -> Optional[str]:
        return getattr(django_settings, 'AWS_SES_RETURN_PATH_ARN', None)

    @property
    def USE_SES_V2(self) -> bool:
        return getattr(django_settings, 'USE_SES_V2', False)

    @property
    def AWS_SES_FROM_EMAIL(self) -> Optional[str]:
        return getattr(django_settings, 'AWS_SES_FROM_EMAIL', None)

    @property
    def TIME_ZONE(self) -> str:
        return django_settings.TIME_ZONE

    @property
    def VERIFY_EVENT_SIGNATURES(self) -> bool:
        return getattr(
            django_settings,
            'AWS_SES_VERIFY_EVENT_SIGNATURES',
            getattr(django_settings, 'AWS_SES_VERIFY_BOUNCE_SIGNATURES', True),
        )

    VERIFY_BOUNCE_SIGNATURES = VERIFY_EVENT_SIGNATURES

    # Domains that are trusted when retrieving the certificate used to sign event messages.
    @property
    def EVENT_CERT_DOMAINS(self) -> List[str]:
        return getattr(
            django_settings,
            'AWS_SNS_EVENT_CERT_TRUSTED_DOMAINS',
            getattr(
                django_settings,
                'AWS_SNS_BOUNCE_CERT_TRUSTED_DOMAINS',
                [
                    'amazonaws.com',
                    'amazon.com',
                ],
            ),
        )

    BOUNCE_CERT_DOMAINS = EVENT_CERT_DOMAINS

    # Blacklists
    @property
    def AWS_SES_ADD_BOUNCE_TO_BLACKLIST(self) -> bool:
        return getattr(django_settings, 'AWS_SES_ADD_BOUNCE_TO_BLACKLIST', False)

    @property
    def AWS_SES_ADD_COMPLAINT_TO_BLACKLIST(self) -> bool:
        return getattr(django_settings, 'AWS_SES_ADD_COMPLAINT_TO_BLACKLIST', False)

    @property
    def AWS_SES_USE_BLACKLIST(self) -> bool:
        return getattr(django_settings, 'AWS_SES_USE_BLACKLIST', False)

    # Inbound
    @property
    def AWS_SES_INBOUND_HANDLER(self) -> str:
        return getattr(django_settings, 'AWS_SES_INBOUND_HANDLER', 'django_ses.inbound.raw_handler')

    # Required if the configured action is set to "Deliver to Amazon S3 bucket"
    # These credentials will be used to access the bucket where the email was stored
    @property
    def AWS_SES_INBOUND_ACCESS_KEY_ID(self) -> str:
        return getattr(django_settings, 'AWS_SES_INBOUND_ACCESS_KEY_ID', '')

    @property
    def AWS_SES_INBOUND_SECRET_ACCESS_KEY(self) -> str:
        return getattr(django_settings, 'AWS_SES_INBOUND_SECRET_ACCESS_KEY', '')

    @property
    def AWS_SES_INBOUND_SESSION_TOKEN(self) -> str:
        return getattr(django_settings, 'AWS_SES_INBOUND_SESSION_TOKEN', '')


settings = SesSettings()



import base64
import binascii
import logging
from email import policy
from email.parser import BytesParser
from typing import TypedDict

import boto3

from django_ses import settings

logger = logging.getLogger(__name__)


class InboundEmail(TypedDict):
    plain_text: str
    html_text: str
    attachments: list
    subject: str
    to: list
    message_id: str
    date: str


class UnprocessableError(Exception):
    pass


class BaseHandler:
  def __init__(self, mail_obj, receipt, raw_message):
      self.mail_obj = mail_obj
      self.receipt = receipt
      self.raw_message = raw_message
      self.action = receipt.get('action', {})

  def check_action_compatibility(self):
      """
      This method should raise UnprocessableError if the received payload from
      SES can't be processed
      """
      pass

  def prepare_content(self, content):
      return content

  def parse_email(self, content) -> InboundEmail:
      parser = BytesParser(policy=policy.default)
      email_message = parser.parsebytes(content)

      plain_text = ''
      html_text = ''
      attachments = []

      for part in email_message.walk():
          content_type = part.get_content_type()
          content_disposition = part.get('Content-Disposition', '')

          # Handle plain text
          if content_type == 'text/plain' and 'attachment' not in content_disposition:
              plain_text = part.get_payload(decode=True).decode(part.get_content_charset() or 'utf-8')

          # Handle HTML
          elif content_type == 'text/html' and 'attachment' not in content_disposition:
              html_text = part.get_payload(decode=True).decode(part.get_content_charset() or 'utf-8')

          # Handle attachments
          elif 'attachment' in content_disposition:
              attachment = {
                  'filename': part.get_filename(),
                  'content_type': content_type,
                  'data': part.get_payload(decode=True)
              }
              attachments.append(attachment)

      common = self.mail_obj.get('commonHeaders', {})

      return InboundEmail(
          plain_text=plain_text,
          html_text=html_text,
          attachments=attachments,
          subject = common.get('subject'),
          to = common.get('to'),
          message_id = common.get('messageId'),
          date = common.get('date'),
      )

  def process(self):
      raise NotImplementedError

  def handle(self, content=None, *args, **kwargs):
      self.check_action_compatibility()

      content = self.prepare_content(content)

      self.email = self.parse_email(content)

      self.process()


class S3Handler(BaseHandler):
  def check_action_compatibility(self):
      action_type = self.action.get('type')
      if action_type != 'S3':
          logger.error(
              f'Received action type ({action_type}) can\'t be handled by this '
              f'handler')
          raise UnprocessableError

  def prepare_content(self, content):
      bucket_name = self.action.get('bucketName')
      object_key = self.action.get('objectKey')

      s3_client = boto3.client(
          's3',
          aws_access_key_id=settings.AWS_SES_INBOUND_ACCESS_KEY_ID,
          aws_secret_access_key=settings.AWS_SES_INBOUND_SECRET_ACCESS_KEY,
          aws_session_token=settings.AWS_SES_INBOUND_SESSION_TOKEN,
      )

      response = s3_client.get_object(Bucket=bucket_name, Key=object_key)
      content = response['Body'].read()

      return content


class SnsHandler(BaseHandler):
  def check_action_compatibility(self):
      action_type = self.action.get('type')
      encoding = self.action.get('encoding')
      if action_type != 'SNS' or encoding != 'BASE64':
          logger.error(
              f'Received action type ({action_type}) or encoding ({encoding}) '
              f'can\'t be handled by this handler')
          raise UnprocessableError

  def prepare_content(self, content):
      try:
          content = base64.b64decode(content, validate=True)
      except binascii.Error:
          logger.error(
              'Content couldn\'t be decoded with base64. Invalid content?')
          raise UnprocessableError

      return content
